#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stscanf_s _stscanf
#   define _stprintf_s(buf, max, format, ...) _sntprintf(buf, max, format, __VA_ARGS__)
#   define _vstprintf_s(buf, max, format, ...) _vsntprintf(buf, max, format, __VA_ARGS__)
#   define memcpy_s(pDst, max, pSrc, n) memcpy(pDst, pSrc, n)
#   define _tcscpy_s(pDstStr, max, pSrcStr) _tcscpy(pDstStr, pSrcStr)
#   define _tcscat_s(pDstStr, max, pStr) _tcscat(pDstStr, pStr)
# endif

#else

/* Linux or VxWorks */
# define _T(c) c
# define _stscanf_s sscanf
# if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
    /* VxWorks 5.x doesn't have snprintf() or vsnprintf() */
#   define _stprintf_s(buf, max, format, ...) sprintf(buf, format, __VA_ARGS__)
#   define _vstprintf_s(buf, max, format, ...) vsprintf(buf, format, __VA_ARGS__)
# else
#   define _stprintf_s(buf, max, format, ...) snprintf(buf, max, format, __VA_ARGS__)
#   define _vstprintf_s(buf, max, format, ...) vsnprintf(buf, max, format, __VA_ARGS__)
# endif
# define memcpy_s(pDst, max, pSrc, n) memcpy(pDst, pSrc, n)
# define _tcscat_s(pDstStr, max, pStr) strcat(pDstStr, pStr)
# define _tcscmp(pStr1, pStr2) strcmp(pStr1, pStr2)
# define _tcscpy_s(pDstStr, max, pSrcStr) strcpy(pDstStr, pSrcStr)
# define _tcslen(pStr) strlen(pStr)
# define _totlower(c) tolower(c)
# define _tprintf printf
# define _istspace(c) isspace(c)

#endif

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include <sdk_common.h>

#include "../../sdk_version.h"

#define NUL (_T('\0'))

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/*
** The environment variable that the sample applications
** attempt to pick up and use to find the base of the ADM-XRC
** SDK and hence the FPGA bitstream files.
*/
#define ENVIRONMENT_VARIABLE _T("ADMXRC_SDK5")

#ifdef WIN32
static const TCHAR nativeSlash = _T('\\');
static const TCHAR nativeSlashStr[] = _T("\\");
static const TCHAR nonNativeSlash = _T('/');
static const TCHAR nonNativeSlashStr[] = _T("/");
#else
static const TCHAR nativeSlash = '/';
static const TCHAR nativeSlashStr[] = "/";
static const TCHAR nonNativeSlash = '\\';
static const TCHAR nonNativeSlashStr[] = "\\";
#endif

static const TCHAR* pTypeString[] = {
  _T("float"),
  _T("float"),
  _T("int"),
  _T("uint"),
  _T("hex"),
  _T("string"),
  _T("boolean"),
  _T("path")
};

#if !defined(_WIN32) || !defined(_MSC_VER)
static void
_tcslwr_s(
  TCHAR* pStr,
  size_t max)
{
  TCHAR c;

  while (max) {
    c = *pStr;
    if (c == NUL) {
      break;
    }
    *pStr++ = tolower(c);
    max--;
  }
}
#endif

#if !defined(_MSC_VER)
static void
_tgetenv_s(
  size_t* pRequiredLength,
  TCHAR* pBuffer,
  size_t max,
  const TCHAR* pVariableName)
{
  TCHAR* pTmp;
  size_t length;

  pTmp = getenv(pVariableName);
  if (NULL == pTmp) {
    length = 0;
  } else {
    length = strlen(pTmp);
  }
  if (NULL != pRequiredLength) {
    /* Not found */
    *pRequiredLength = length + 1;
  }
  if (NULL != pBuffer) {
    if (max) {
      max--; /* subtract 1 for NUL terminator */
      if (length > max) {
        length = max;
      }
      if (length) {
        memcpy(pBuffer, pTmp, length * sizeof(TCHAR));
      }
      pBuffer[length] = '\0';
    }
  }
}
#endif

static const TCHAR*
getTypeString(
  int type)
{
  if (type < 0 || type >= ARRAY_LENGTH(pTypeString)) {
    return _T("?");
  } else {
    return pTypeString[type];
  }
}

static void
set_string(
  TCHAR** ppDst,
  const TCHAR* pSrc)
{
  TCHAR* pStr = NULL;
  size_t n;

  assert(NULL != ppDst);
  pStr = *ppDst;
  if (pStr != NULL) {
    free(pStr);
    *ppDst = NULL;
  }
  if (pSrc != NULL) {
    n = _tcslen(pSrc) + 1;
    pStr = (TCHAR*)malloc(sizeof(TCHAR) * n);
    assert(NULL != pStr);
    if (NULL != pStr) {
      pStr[0] = _T('\0');
      _tcscpy_s(pStr, n, pSrc);
    }
    *ppDst = pStr;
  }
}

static void
fatal(
  int code,
  TCHAR* fmt, ...)
{
  TCHAR buffer[1024];
  va_list ap;
  size_t n;

  va_start(ap, fmt);
#if defined(_MSC_VER)
  n = _vsntprintf_s(buffer, sizeof(buffer), _TRUNCATE, fmt, ap);
#else
  n = vsnprintf(buffer, sizeof(buffer), fmt, ap);
#endif
  va_end(ap);

  _tprintf(_T("%s"), buffer);

  exit(code);
}

static void
convert_slashes(
  TCHAR* pStr)
{
  while (*pStr != _T('\0')) {
    if (*pStr == nonNativeSlash) {
      *pStr = nativeSlash;
    }
    pStr++;
  }
}

static void
strip_leading_space(
  TCHAR* pStr)
{
  TCHAR* p = pStr;
  TCHAR* q = pStr;

  while (*p != _T('\0') && _istspace(*p)) {
    p++;
  }
  while (*p != _T('\0')) {
    *q++ = *p++;
  }
  *q = 0;
}

static void
strip_trailing_space(
  TCHAR* pStr)
{
  size_t i, n;

  n = _tcslen(pStr);
  if (n == 0) {
    return;
  }

  for (i = n - 1; i >= 0; i--) {
    if (!_istspace(pStr[i])) {
      i++;
      break;
    }
  }
  pStr[i] = _T('\0');
}

static void
strip_trailing_slash(
  TCHAR* pStr)
{
  size_t n;

  n = _tcslen(pStr);
  if (n) {
    if (pStr[n - 1] == nativeSlash) {
      if (n != 1) {
        pStr[n - 1] = _T('\0');
      }
    }
  }
}

static void
tidy_path(
  TCHAR* pStr)
{
  convert_slashes(pStr);
  strip_leading_space(pStr);
  strip_trailing_space(pStr);
  strip_trailing_slash(pStr);
}

static int
set_value(
  Value* pValue,
  OptionType type,
  const TCHAR* pStr)
{
  TCHAR buff[6];
  union {
    float floatVal;
    long intVal;
    unsigned long uintVal;
    bool_t booleanVal;
  } val;

  if (type == Option_float) {
    _stscanf_s(pStr, _T("%f"), &val.floatVal);
    pValue->floatVal = val.floatVal;
  } else if (type == Option_double) {
    _stscanf_s(pStr, _T("%f"), &val.floatVal);
    pValue->doubleVal = (double)val.floatVal;
  } else if (type == Option_int) {
    _stscanf_s(pStr, _T("%ld"), &val.intVal);
    pValue->intVal = val.intVal;
  } else if (type == Option_uint) {
    if (_tcslen(pStr) >= 2 && pStr[0] == _T('0') && _totlower(pStr[1]) == _T('x')) {
      _stscanf_s(pStr + 2, _T("%lx"), &val.uintVal);
    } else {
      _stscanf_s(pStr, _T("%lu"), &val.uintVal);
    }
    pValue->uintVal = val.uintVal;
  } else if (type == Option_hex) {
    _stscanf_s(pStr, _T("%lx"), &val.uintVal);
    pValue->hexVal = val.uintVal;
  } else if (type == Option_string) {
    set_string(&pValue->stringVal, pStr);
  } else if (type == Option_boolean) {
    _tcscpy_s(buff, ARRAY_LENGTH(buff), pStr);
    _tcslwr_s(buff, ARRAY_LENGTH(buff));
    if (_tcscmp(buff, _T("true")) == 0) {
      pValue->booleanVal = TRUE;
    } else if (_tcscmp(buff, _T("false")) == 0) {
      pValue->booleanVal = FALSE;
    } else {
      return -1;
    }
  } else if (type == Option_path) {
    set_string(&pValue->pathVal, pStr);
    tidy_path(pValue->pathVal);
  } else {
    _tprintf(_T("*** internal error in args.c:set_value()\n"));
    return -1;
  }

  return 0;
}

static TCHAR*
default_to_string(
  const Option* pOption)
{
  static TCHAR buffer[128];

  buffer[0] = _T('\0');
  switch(pOption->type) {
  case Option_float:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%f"), (float)pOption->def.floatVal);
    break;

  case Option_double:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%lf"), pOption->def.doubleVal);
    break;

  case Option_int:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%ld"), pOption->def.intVal);
    break;

  case Option_uint:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%lu"), pOption->def.uintVal);
    break;

  case Option_hex:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("0x%lX"), pOption->def.hexVal);
    break;

  case Option_string:
    if (pOption->def.stringVal != NULL) {
      _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%s"), pOption->def.stringVal);
    } else {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("(null)"));
    }
    break;

  case Option_boolean:
    if (pOption->def.booleanVal) {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("TRUE"));
    } else {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("FALSE"));
    }
    break;

  case Option_path:
    if (pOption->def.pathVal != NULL) {
      _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%s"), pOption->def.pathVal);
    } else {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("(null)"));
    }
    break;
  }

  return buffer;
}

static TCHAR*
value_to_string(
  const Value* pValue,
  OptionType  type)
{
  static TCHAR buffer[128];

  switch(type) {
  case Option_float:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%f"), (float)pValue->floatVal);
    break;

  case Option_double:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%lf"), pValue->doubleVal);
    break;

  case Option_int:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%ld"), pValue->intVal);
    break;

  case Option_uint:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%lu"), pValue->uintVal);
    break;

  case Option_hex:
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("0x%lX"), pValue->hexVal);
    break;

  case Option_string:
    if (pValue->stringVal != NULL) {
      _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%s"), pValue->stringVal);
    } else {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("(null)"));
    }
    break;

  case Option_boolean:
    if (pValue->booleanVal) {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("TRUE"));
    } else {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("FALSE"));
    }
    break;

  case Option_path:
    if (pValue->pathVal != NULL) {
      _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%s"), pValue->pathVal);
    } else {
      _tcscpy_s(buffer, ARRAY_LENGTH(buffer), _T("(null)"));
    }
    break;
  }

  return buffer;
}

static TCHAR*
option_to_string(
  Option* option)
{
  static TCHAR* pValStr;
  static TCHAR buffer[128];

  pValStr = value_to_string(&option->value, option->type);
  if (option->type == Option_boolean) {
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("%c%s %s"), option->value.booleanVal ? _T('+') : _T('-'), option->key, pValStr);
  } else {
    _stprintf_s(buffer, ARRAY_LENGTH(buffer), _T("-%s %s"), option->key, pValStr);
  }

  return buffer;
}

static void
set_option_default(
  Option* pOption)
{
  switch (pOption->type) {
  case Option_float:
    pOption->value.floatVal = pOption->def.floatVal;
    break;

  case Option_double:
    pOption->value.doubleVal = pOption->def.doubleVal;
    break;

  case Option_int:
    pOption->value.intVal = pOption->def.intVal;
    break;

  case Option_uint:
    pOption->value.uintVal = pOption->def.uintVal;
    break;

  case Option_hex:
    pOption->value.hexVal = pOption->def.hexVal;
    break;

  case Option_string:
    pOption->value.stringVal = NULL;
    set_string(&pOption->value.stringVal, pOption->def.stringVal);
    break;

  case Option_boolean:
    pOption->value.booleanVal = pOption->def.booleanVal;
    break;

  case Option_path:
    pOption->value.pathVal = NULL;
    set_string(&pOption->value.pathVal, pOption->def.pathVal);
    break;
  }
  pOption->specified = FALSE;
}

static void
set_option(
  Option* pOption,
  const TCHAR* pStr)
{
  union {
    float floatVal;
    long intVal;
    unsigned long uintVal;
    bool_t booleanVal;
  } value;

  switch (pOption->type) {
  case Option_float:
    _stscanf_s(pStr, _T("%f"), &value.floatVal);
    pOption->value.floatVal = value.floatVal;
    break;

  case Option_double:
    _stscanf_s(pStr, _T("%f"), &value.floatVal);
    pOption->value.doubleVal = (double)value.floatVal;
    break;

  case Option_int:
    _stscanf_s(pStr, _T("%ld"), &value.intVal);
    pOption->value.intVal = value.intVal;
    break;

  case Option_uint:
    if (_tcslen(pStr) >= 2 && pStr[0] == _T('0') && pStr[1] == _T('x')) {
      _stscanf_s(pStr + 2, _T("%lx"), &value.intVal);
    } else {
      _stscanf_s(pStr, _T("%lu"), &value.intVal);
    }
    pOption->value.uintVal = value.uintVal;
    break;

  case Option_hex:
    _stscanf_s(pStr, _T("%lx"), &value.uintVal);
    pOption->value.hexVal = value.uintVal;
    break;

  case Option_string:
    set_string(&pOption->value.stringVal, pStr);
    break;

  case Option_path:
    set_string(&pOption->value.pathVal, pStr);
    tidy_path(pOption->value.pathVal);
    break;

  default:
    assert(FALSE);
    break;
  }
  pOption->specified = TRUE;
}

static Option*
find_option(
  const TCHAR* pKeyStr,
  Option* pOptions,
  int n)
{
  int i;

  for (i = 0; i < n; i++) {
    if (_tcscmp(pOptions[i].key, pKeyStr) == 0) {
      return &pOptions[i];
    }
  }

  return NULL;
}

void
show_usage(
  const TCHAR* pAppDescStr,
  const TCHAR* pAppNameStr,
  Arguments* pArgs)
{
  int i;
  unsigned int version[4] = { ADMXRC_SDK_VERSION_NUM };

  _tprintf(_T("%s V%u.%u.%u Copyright (C) Alpha Data 2000-2011\n"), pAppDescStr, version[0], version[1], version[2]);
  _tprintf(_T("Usage: %s [option ...]"), pAppNameStr);
  for (i = 0; i < pArgs->nParamType; i++) {
    TCHAR* a, * b;

    if (i < pArgs->minNParam) {
      a = _T("");
      b = _T("");
    } else {
      a = _T("[");
      b = _T("]");
    }
    _tprintf(_T(" %s%s%s"), a, pArgs->paramInfo[i], b);
  }
  if (pArgs->nParamType) {
    _tprintf(_T("\n"));
  }
  _tprintf(_T("\n"));
  for (i = 0; i < pArgs->nParamType; i++) {
    _tprintf(_T("%10s = %s\n"), pArgs->paramInfo[i], pArgs->paramMeaning[i]);
  }
  _tprintf(_T("\n"));
  _tprintf(_T("%12s  %7s  %12s  %s\n\n"), _T("Option"), _T("Type"), _T("Default"), _T("Meaning"));
#ifdef WIN32
  _tprintf(_T("%12s  %7s  %12s  %s\n"), _T("/h /? /help"), _T(""), _T(""), _T("Show this help"));
  _tprintf(_T("%12s  %7s  %12s  %s\n"), _T("/sdk"), "string", _T(""), _T("Specify base directory of ADM-XRC SDK"));
  _tprintf(_T("%12s  %7s  %12s  %s\n"), _T("/bitdir"), "string", _T(""), _T("Specify bitstream directory"));
#else
  _tprintf(_T("%12s  %7s  %12s  %s\n"), _T("-h -? -help"), _T(""), _T(""), _T("Show this help"));
  _tprintf(_T("%12s  %7s  %12s  %s\n"), _T("-sdk"), "string", _T(""), _T("Specify base directory of ADM-XRC SDK"));
  _tprintf(_T("%12s  %7s  %12s  %s\n"), _T("-bitdir"), "string", _T(""), _T("Specify bitstream directory"));
#endif
  for (i = 0; i < pArgs->nOption; i++) {
    TCHAR buff[17];

    if (pArgs->option[i].type == Option_boolean) {
      _tcscpy_s(buff, ARRAY_LENGTH(buff), _T("-/+"));
    } else {
      _tcscpy_s(buff, ARRAY_LENGTH(buff), _T("/"));
    }
    _tcscat_s(buff, ARRAY_LENGTH(buff), pArgs->option[i].key);
    _tprintf(_T("%12s  %7s  %12s  %s\n"),
      buff,
      getTypeString(pArgs->option[i].type),
      default_to_string(&pArgs->option[i]),
      pArgs->option[i].meaning);
  }
}

void
dump_command_line(
  const Arguments* pArgs)
{
  TCHAR* buffer;
  int   i, j;

  _tprintf(_T("Command line: %s"), pArgs->appName);
  for (i = 0; i < pArgs->nOption; i++) {
    if (pArgs->option[i].specified) {
      buffer = option_to_string(&pArgs->option[i]);
      _tprintf(_T(" %s"), buffer);
    }
  }
  j = pArgs->nParam;
  if (j > pArgs->nParamType) {
    j = pArgs->nParamType;
  }
  for (i = 0; i < j; i++) {
    buffer = value_to_string(&pArgs->param[i], pArgs->paramType[i]);
    _tprintf(_T(" %s"), buffer);
  }
  for (; i < pArgs->nParam; i++) {
    buffer = value_to_string(&pArgs->param[i], Option_string);
    _tprintf(_T(" %s"), buffer);
  }
  _tprintf(_T("\n"));
}

void
dump_options(
  const Arguments* pArgs)
{
  int i;

  _tprintf(_T("Dump of all options:\n"));
  for (i = 0; i < pArgs->nOption; i++) {
    TCHAR* pBuffer = value_to_string(&pArgs->option[i].value, pArgs->option[i].type);
    _tprintf(_T("%12s %8s %8s\n"), pArgs->option[i].key, getTypeString(pArgs->option[i].type), pBuffer);
  }
}

int
parse_command_line(
  const TCHAR* pAppDescStr,
  const TCHAR* pAppNameStr,
  int argc,
  TCHAR* argv[],
  Arguments* pArgs)
{
  int i, j;
  size_t n;

  if (pAppNameStr == NULL) {
    fatal(-1, _T("parse_command_line(): argument 'pAppNameStr' is NULL (must not be NULL)\n"));
  }
  if (pArgs == NULL) {
    fatal(-1, _T("parse_command_line(): argument 'args' is NULL (must not be NULL)\n"));
  }

  pArgs->appName = NULL;
  pArgs->sdkPath = NULL;
  pArgs->bitPath = NULL;
  pArgs->bitFile = NULL;
  set_string(&pArgs->appName, pAppNameStr);

  for (i = 0; i < pArgs->nOption; i++) {
    set_option_default(&pArgs->option[i]);
  }

  _tgetenv_s(&n, NULL, 0, ENVIRONMENT_VARIABLE);
  if (n > 0) {
    size_t nActual;
    
    pArgs->sdkPath = (TCHAR*)malloc(n * sizeof(TCHAR));
    _tgetenv_s(&nActual, pArgs->sdkPath, n, ENVIRONMENT_VARIABLE);
    assert(nActual == n);
  }

  // Parse the options
  for (i = 1; i < argc; i++) {
    TCHAR tag = argv[i][0];

    if ((tag == _T('-')) || (tag == _T('/')) || (tag == _T('+'))) {
      if ((_tcscmp(argv[i] + 1, _T("h")) == 0) ||
          (_tcscmp(argv[i] + 1, _T("help")) == 0) ||
          (_tcscmp(argv[i] + 1, _T("?")) == 0)) {
        show_usage(pAppDescStr, pAppNameStr, pArgs);
        exit(0);
      } else if (_tcscmp(argv[i] + 1, _T("sdk")) == 0) {
        i++;
        if (i == argc) {
          fatal(-1, _T("/sdk option requires parameter: path to ADM-XRC SDK root\n"));
        }
        set_string(&pArgs->sdkPath, argv[i]);
        tidy_path(pArgs->sdkPath);
      } else if (_tcscmp(argv[i] + 1, _T("bitdir")) == 0) {
        i++;
        if (i == argc) {
          fatal(-1, _T("/bitdir option requires parameter: path to bitstream directory\n"));
        }
        set_string(&pArgs->bitPath, argv[i]);
        tidy_path(pArgs->bitPath);
      } else {
        int narg;
        Option* pOption;

        pOption = find_option(argv[i] + 1, pArgs->option, pArgs->nOption);
        if (pOption == NULL) {
          fatal(-1, _T("Unknown option: %s\n"), argv[i]);
        } else {
          narg = (pOption->type == Option_boolean) ? 0 : 1;
          if (i + narg + 1 > argc) {
            fatal(-1, _T("%s option requires parameter: %s\n"), argv[i], pOption->meaning);
          }
          if (pOption->type == Option_boolean) {
            if (tag == _T('-')) {
              pOption->value.booleanVal = FALSE;
            } else if (tag == _T('+')) {
              pOption->value.booleanVal = TRUE;
            } else {
              fatal(-1, _T("Boolean option %s should be specified as -%s or +%s\n"), pOption->key, pOption->key, pOption->key);
            }
          } else {
            i++;
            set_option(pOption, argv[i]);
          }
        }
      }
    } else {
      break;
    }
  }

  pArgs->nParam = argc - i;
  if (pArgs->nParam == 0) {
    pArgs->param = NULL;
  } else {
    pArgs->param = malloc(sizeof(Value) * pArgs->nParam);
    j = 0;
    for (; i < argc; i++) {
      if (j < pArgs->nParamType) {
        if (pArgs->paramType[j] == Option_string) {
          pArgs->param[j].stringVal = NULL;
        } else if (pArgs->paramType[j] == Option_path) {
          pArgs->param[j].pathVal = NULL;
        }
        if (set_value(&pArgs->param[j], pArgs->paramType[j], argv[i])) {
          fatal(-1, _T("Value of %s parameter %d is invalid: %s\n"), getTypeString(pArgs->paramType[i]), i + 1, argv[i]);
        }
      } else {
        set_value(&pArgs->param[j], Option_string, argv[i]);
      }
      j++;
    }
  }

  if (pArgs->minNParam > pArgs->nParam) {
    _tprintf(_T("Too few parameters supplied. Use %s /? to get help.\n"), pArgs->appName);
    exit(-1);
  }

  if (pArgs->bitPath == NULL && pArgs->sdkPath != NULL) {
    const TCHAR* bitDir = _T("bit");
    TCHAR* pBitpathStr;
    size_t n;

    n = _tcslen(pArgs->sdkPath) + 1 + _tcslen(bitDir) + 1;
    pBitpathStr = (TCHAR*)malloc(n * sizeof(TCHAR));
    if (pBitpathStr == NULL) {
      fatal(-1, _T("parse_command_line(): failed to allocate 'pBitpathStr'\n"));
    }
    _tcscpy_s(pBitpathStr, n, pArgs->sdkPath);
    _tcscat_s(pBitpathStr, n, nativeSlashStr);
    _tcscat_s(pBitpathStr, n, bitDir);
    set_string(&pArgs->bitPath, pBitpathStr);
    free(pBitpathStr);
  }

  return 0;
}

const TCHAR*
get_fpga_filename(
  const ADMXRC2_CARD_INFO* pCardInfo,
  const TCHAR* pDesignStr,
  Arguments* pArgs)
{
  return get_fpga_filename_dir(pCardInfo, pDesignStr, pDesignStr, pArgs);
}

const TCHAR*
get_fpga_filename_dir(
  const ADMXRC2_CARD_INFO* pCardInfo,
  const TCHAR* pDirStr,
  const TCHAR* pDesignStr,
  Arguments* pArgs)
{
  const TCHAR ext[] = _T(".bit");
  TCHAR* pBoardStr;
  TCHAR* pDeviceStr;
  TCHAR* pBitfileStr;

  if (pDesignStr == NULL) {
    fatal(-1, _T("get_fpga_filename(): argument 'pDesignStr' is NULL (must not be NULL)\n"));
  }

  switch(pCardInfo->BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
    pBoardStr = _T("xrc");
    break;

  case ADMXRC2_BOARD_ADMXRC_P:
    pBoardStr = _T("xrcp");
    break;

  case ADMXRC2_BOARD_ADMXRC2_LITE:
    pBoardStr = _T("xrc2l");
    break;

  case ADMXRC2_BOARD_ADMXRC2:
    pBoardStr = _T("xrc2");
    break;

  case ADMXRC2_BOARD_ADMXPL:
    pBoardStr = _T("xpl");
    break;

  case ADMXRC2_BOARD_ADMXP:
    pBoardStr = _T("xp");
    break;

  case ADMXRC2_BOARD_ADPWRC2:
    pBoardStr = _T("wrc2");
    break;

  case ADMXRC2_BOARD_ADPDRC2:
    pBoardStr = _T("drc2");
    break;

  case ADMXRC2_BOARD_ADPXPI:
    pBoardStr = _T("xpi");
    break;

  case ADMXRC2_BOARD_ADMXRC4LS:
    pBoardStr = _T("xrc4ls");
    break;

  case ADMXRC2_BOARD_ADMXRC4LX:
    pBoardStr = _T("xrc4lx");
    break;

  case ADMXRC2_BOARD_ADMXRC4SX:
    pBoardStr = _T("xrc4sx");
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
    pBoardStr = _T("xrc4fx");
    break;

  case ADMXRC2_BOARD_ADPEXRC4FX:
    pBoardStr = _T("xrce4fx");
    break;

  case ADMXRC2_BOARD_ADMXRC5LX:
    pBoardStr = _T("xrc5lx");
    break;

  case ADMXRC2_BOARD_ADMXRC5T1:
    pBoardStr = _T("xrc5t1");
    break;

  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    pBoardStr = _T("xrc5t2");
    break;

  case ADMXRC2_BOARD_ADCPXRC4LX:
    pBoardStr = _T("xrcp4lx");
    break;

  case ADMXRC2_BOARD_ADMAMC5A2:
    pBoardStr = _T("amc5a2");
    break;

  case ADMXRC2_BOARD_ADMXRC5TZ:
    pBoardStr = _T("xrc5tz");
    break;

  case ADMXRC2_BOARD_ADMXRC5TDA1:
    pBoardStr = _T("xrc5tda1");
    break;

  case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
    pBoardStr = _T("xrc5t2advcc1");
    break;

  case ADMXRC2_BOARD_ADMPCIE6S1:
    pBoardStr = _T("pcie6s1");
    break;

  default:
    fatal(-1, _T("Unknown board type: %d\n"), pCardInfo->BoardType);
  }

  switch(pCardInfo->FPGAType) {
  case ADMXRC2_FPGA_V405E:
    pDeviceStr = _T("v405e");
    break;

  case ADMXRC2_FPGA_V812E:
    pDeviceStr = _T("v812e");
    break;

  case ADMXRC2_FPGA_V3200E:
    pDeviceStr = _T("v3200e");
    break;

  case ADMXRC2_FPGA_V2000E:
    pDeviceStr = _T("v2000e");
    break;

  case ADMXRC2_FPGA_V1600E:
    pDeviceStr = _T("v1600e");
    break;

  case ADMXRC2_FPGA_V1000E:
    pDeviceStr = _T("v1000e");
    break;

  case ADMXRC2_FPGA_V1000:
    pDeviceStr = _T("v1000");
    break;

  case ADMXRC2_FPGA_V800:
    pDeviceStr = _T("v800");
    break;

  case ADMXRC2_FPGA_V600:
    pDeviceStr = _T("v600");
    break;

  case ADMXRC2_FPGA_V400:
    pDeviceStr = _T("v400");
    break;

  case ADMXRC2_FPGA_2V1000:
    pDeviceStr = _T("2v1000");
    break;

  case ADMXRC2_FPGA_2V1500:
    pDeviceStr = _T("2v1500");
    break;

  case ADMXRC2_FPGA_2V2000:
    pDeviceStr = _T("2v2000");
    break;

  case ADMXRC2_FPGA_2V3000:
    pDeviceStr = _T("2v3000");
    break;

  case ADMXRC2_FPGA_2V4000:
    pDeviceStr = _T("2v4000");
    break;

  case ADMXRC2_FPGA_2V6000:
    pDeviceStr = _T("2v6000");
    break;

  case ADMXRC2_FPGA_2V8000:
    pDeviceStr = _T("2v8000");
    break;

  case ADMXRC2_FPGA_2V10000:
    pDeviceStr = _T("2v10000");
    break;

  case ADMXRC2_FPGA_2VP2:
    pDeviceStr = _T("2vp2");
    break;

  case ADMXRC2_FPGA_2VP4:
    pDeviceStr = _T("2vp4");
    break;

  case ADMXRC2_FPGA_2VP7:
    pDeviceStr = _T("2vp7");
    break;

  case ADMXRC2_FPGA_2VP20:
    pDeviceStr = _T("2vp20");
    break;

  case ADMXRC2_FPGA_2VP30:
    pDeviceStr = _T("2vp30");
    break;

  case ADMXRC2_FPGA_2VP40:
    pDeviceStr = _T("2vp40");
    break;

  case ADMXRC2_FPGA_2VP50:
    pDeviceStr = _T("2vp50");
    break;

  case ADMXRC2_FPGA_2VP100:
    pDeviceStr = _T("2vp100");
    break;

  case ADMXRC2_FPGA_2VP125:
    pDeviceStr = _T("2vp125");
    break;

  case ADMXRC2_FPGA_2VP70:
    pDeviceStr = _T("2vp70");
    break;

  case ADMXRC2_FPGA_4VLX15:
    pDeviceStr = _T("4vlx15");
    break;

  case ADMXRC2_FPGA_4VLX25:
    pDeviceStr = _T("4vlx25");
    break;

  case ADMXRC2_FPGA_4VLX40:
    pDeviceStr = _T("4vlx40");
    break;

  case ADMXRC2_FPGA_4VLX60:
    pDeviceStr = _T("4vlx60");
    break;

  case ADMXRC2_FPGA_4VLX80:
    pDeviceStr = _T("4vlx80");
    break;

  case ADMXRC2_FPGA_4VLX100:
    pDeviceStr = _T("4vlx100");
    break;

  case ADMXRC2_FPGA_4VLX160:
    pDeviceStr = _T("4vlx160");
    break;

  case ADMXRC2_FPGA_4VLX200:
    pDeviceStr = _T("4vlx200");
    break;

  case ADMXRC2_FPGA_4VSX25:
    pDeviceStr = _T("4vsx25");
    break;

  case ADMXRC2_FPGA_4VSX35:
    pDeviceStr = _T("4vsx35");
    break;

  case ADMXRC2_FPGA_4VSX55:
    pDeviceStr = _T("4vsx55");
    break;

  case ADMXRC2_FPGA_4VFX12:
    pDeviceStr = _T("4vfx12");
    break;

  case ADMXRC2_FPGA_4VFX20:
    pDeviceStr = _T("4vfx20");
    break;

  case ADMXRC2_FPGA_4VFX40:
    pDeviceStr = _T("4vfx40");
    break;
  case ADMXRC2_FPGA_4VFX60:
    pDeviceStr = _T("4vfx60");
    break;

  case ADMXRC2_FPGA_4VFX100:
    pDeviceStr = _T("4vfx100");
    break;

  case ADMXRC2_FPGA_4VFX140:
    pDeviceStr = _T("4vfx140");
    break;

  case ADMXRC2_FPGA_5VLX30:
    pDeviceStr = _T("5vlx30");
    break;

  case ADMXRC2_FPGA_5VLX50:
    pDeviceStr = _T("5vlx50");
    break;

  case ADMXRC2_FPGA_5VLX85:
    pDeviceStr = _T("5vlx85");
    break;

  case ADMXRC2_FPGA_5VLX110:
    pDeviceStr = _T("5vlx110");
    break;

  case ADMXRC2_FPGA_5VLX220:
    pDeviceStr = _T("5vlx220");
    break;

  case ADMXRC2_FPGA_5VLX330:
    pDeviceStr = _T("5vlx330");
    break;

  case ADMXRC2_FPGA_5VLX155:
    pDeviceStr = _T("5vlx155");
    break;

  case ADMXRC2_FPGA_5VLX30T:
    pDeviceStr = _T("5vlx30t");
    break;

  case ADMXRC2_FPGA_5VLX50T:
    pDeviceStr = _T("5vlx50t");
    break;

  case ADMXRC2_FPGA_5VLX85T:
    pDeviceStr = _T("5vlx85t");
    break;

  case ADMXRC2_FPGA_5VLX110T:
    pDeviceStr = _T("5vlx110t");
    break;

  case ADMXRC2_FPGA_5VLX330T:
    pDeviceStr = _T("5vlx330t");
    break;

  case ADMXRC2_FPGA_5VLX220T:
    pDeviceStr = _T("5vlx220t");
    break;

  case ADMXRC2_FPGA_5VLX155T:
    pDeviceStr = _T("5vlx155t");
    break;

  case ADMXRC2_FPGA_5VSX35T:
    pDeviceStr = _T("5vsx35t");
    break;

  case ADMXRC2_FPGA_5VSX50T:
    pDeviceStr = _T("5vsx50t");
    break;

  case ADMXRC2_FPGA_5VSX95T:
    pDeviceStr = _T("5vsx95t");
    break;

  case ADMXRC2_FPGA_5VSX240T:
    pDeviceStr = _T("5vsx240t");
    break;

  case ADMXRC2_FPGA_5VFX100T:
    pDeviceStr = _T("5vfx100t");
    break;

  case ADMXRC2_FPGA_5VFX130T:
    pDeviceStr = _T("5vfx130t");
    break;

  case ADMXRC2_FPGA_5VFX200T:
    pDeviceStr = _T("5vfx200t");
    break;

  case ADMXRC2_FPGA_5VFX30T:
    pDeviceStr = _T("5vfx30t");
    break;

  case ADMXRC2_FPGA_5VFX70T:
    pDeviceStr = _T("5vfx70t");
    break;

  case ADMXRC2_FPGA_6SLX4:
    pDeviceStr = _T("6slx4");
    break;

  case ADMXRC2_FPGA_6SLX9:
    pDeviceStr = _T("6slx9");
    break;

  case ADMXRC2_FPGA_6SLX16:
    pDeviceStr = _T("6slx16");
    break;

  case ADMXRC2_FPGA_6SLX25:
    pDeviceStr = _T("6slx25");
    break;

  case ADMXRC2_FPGA_6SLX45:
    pDeviceStr = _T("6slx45");
    break;

  case ADMXRC2_FPGA_6SLX75:
    pDeviceStr = _T("6slx75");
    break;

  case ADMXRC2_FPGA_6SLX100:
    pDeviceStr = _T("6slx100");
    break;

  case ADMXRC2_FPGA_6SLX150:
    pDeviceStr = _T("6slx150");
    break;

  case ADMXRC2_FPGA_6SLX25T:
    pDeviceStr = _T("6slx25t");
    break;

  case ADMXRC2_FPGA_6SLX45T:
    pDeviceStr = _T("6slx45t");
    break;

  case ADMXRC2_FPGA_6SLX75T:
    pDeviceStr = _T("6slx75t");
    break;

  case ADMXRC2_FPGA_6SLX100T:
    pDeviceStr = _T("6slx100t");
    break;

  case ADMXRC2_FPGA_6SLX150T:
    pDeviceStr = _T("6slx150t");
    break;

  default:
    fatal(-1, _T("Unknown FPGA type on card: %d\n"), pCardInfo->FPGAType);
  }

  if (pArgs->bitPath != NULL) {
    size_t n = _tcslen(pArgs->bitPath) + 1 + _tcslen(pDirStr) + 1 + _tcslen(pDesignStr) + 1 + _tcslen(pBoardStr) + 1 + _tcslen(pDeviceStr) +  _tcslen(ext) + 1;

    pBitfileStr = (TCHAR*)malloc(n * sizeof(TCHAR));
    if (pBitfileStr == NULL) {
      fatal(-1, _T("get_fpga_filename(): failed to allocate 'bitfile'\n"));
    }
    _tcscpy_s(pBitfileStr, n, pArgs->bitPath);
    _tcscat_s(pBitfileStr, n, nativeSlashStr);
    _tcscat_s(pBitfileStr, n, pDirStr);
    _tcscat_s(pBitfileStr, n, nativeSlashStr);
    _tcscat_s(pBitfileStr, n, pDesignStr);
    _tcscat_s(pBitfileStr, n, _T("-"));
    _tcscat_s(pBitfileStr, n, pBoardStr);
    _tcscat_s(pBitfileStr, n, _T("-"));
    _tcscat_s(pBitfileStr, n, pDeviceStr);
    _tcscat_s(pBitfileStr, n, ext);
    set_string(&pArgs->bitFile, pBitfileStr);
    free(pBitfileStr);
  } else {
    fatal(
      -1,
      _T("The path to the FPGA bitstream directory cannot be determined via the ")
      _T("%s environment variable and has not been specified via the ")
      _T("via the /sdk or /bitdir switches. Please correct this and ")
      _T("try again.\n"),
      ENVIRONMENT_VARIABLE);
  }

  return pArgs->bitFile;
}
