#ifndef _ADATA_SDK_COMMON_H
#define _ADATA_SDK_COMMON_H

#include <admxrc2.h>

#if defined(_WIN32)

typedef BOOL bool_t;
typedef INT8 int8_t;
typedef INT16 int16_t;
typedef INT32 int32_t;
typedef INT64 int64_t;
typedef UINT8 uint8_t;
typedef UINT16 uint16_t;
typedef UINT32 uint32_t;
typedef UINT64 uint64_t;

#else

# ifndef FALSE
#   define FALSE (0)
# endif
# ifndef TRUE
#   define TRUE (1)
# endif

typedef int bool_t;
typedef char TCHAR;

#endif

typedef enum _OptionType {
  Option_float   = 0x0,
  Option_double  = 0x1,
  Option_int     = 0x2,
  Option_uint    = 0x3,
  Option_hex     = 0x4,
  Option_string  = 0x5,
  Option_boolean = 0x6,
  Option_path    = 0x7
} OptionType;

typedef union _Value {
  float floatVal;
  double doubleVal;
  long intVal;
  unsigned long uintVal;
  unsigned long hexVal;
  TCHAR* stringVal;
  bool_t booleanVal;
  TCHAR* pathVal;
} Value;

typedef struct _Option {
  /* Must be initialised before call */
  const TCHAR* key;
  int type;
  const TCHAR* meaning;
  union _Value def;
  /* Filled in on return */
  bool_t specified;
  union _Value value;
  int nParam;
  union _Value* param;
} Option;

typedef struct _Arguments {
  /* Initialised before call */
  int nOption;
  OptionType* paramType;
  TCHAR** paramInfo;
  TCHAR** paramMeaning;
  int nParamType;
  int minNParam;
  /* Partially initialised before call */
  Option* option;
  /* Filled in on return */
  int nParam;
  Value* param;
  TCHAR* appName;
  TCHAR* sdkPath;
  TCHAR* bitPath;
  TCHAR* bitFile;
} Arguments;

int
parse_command_line(
  const TCHAR* pAppDescStr,
  const TCHAR* pAppNameStr,
  int argc,
  TCHAR* argv[],
  Arguments* pArgs);

const TCHAR*
get_fpga_filename(
  const ADMXRC2_CARD_INFO* pCardInfo,
  const TCHAR* pDesignStr,
  Arguments* pArgs);

const TCHAR*
get_fpga_filename_dir(
  const ADMXRC2_CARD_INFO* cardInfo,
  const TCHAR* pDirStr,
  const TCHAR* pDesignStr,
  Arguments* args);

void
dump_options(
  const Arguments* pArgs);

void
dump_command_line(
  const Arguments* pArgs);

#endif
