#ifndef _ADATA_PLATFORM_TIME_H
#define _ADATA_PLATFORM_TIME_H

#include <windows.h>
#include <assert.h>

typedef DWORD admxrc2_time_t;

#define timeGet() GetTickCount()

static const admxrc2_time_t timeZero = 0U;

__inline static admxrc2_time_t
timeMake(
  long sec,
  long usec,
  BOOL bRoundUp)
{
  if (bRoundUp) {
    return sec * 1000 + (usec + 999) / 1000;
  } else {
    return sec * 1000 + usec / 1000;
  }
}

#define timeAfter(t1, t2) ((t1) > (t2))

#define timeBefore(t1, t2) ((t1) < (t2))

#define timeEqual(t1, t2) ((t1) == (t2))

#define timeAdd(t1, t2) ((t1) + (t2))

#define timeSubtract(t1, t2) ((t1) - (t2))

#define timeToDouble(t) ((double)(t) / 1000.0)

#define timeToFloat(t) ((float)(t) / 1000.0)

#define doubleToTime(d) ((DWORD)(d * 1000.0))

#define floatToTime(f) ((DWORD)(d * 1000.0f))

#define sleepMs(ms) Sleep(ms)

#if defined(ADMXRC3_IMPORT_DELAY_US)

static LARGE_INTEGER admxrc3_hpcTickRate;

__inline static void
initDelayUs(
  void)
{
  admxrc3_hpcTickRate.QuadPart = 0;
  QueryPerformanceFrequency(&admxrc3_hpcTickRate);
  assert(0 != admxrc3_hpcTickRate.QuadPart);
}

__inline static void
delayUs(
  unsigned long us)
{
  LARGE_INTEGER expire, now;

  QueryPerformanceCounter(&expire);
  expire.QuadPart += admxrc3_hpcTickRate.QuadPart * us / 1000000;
  do {
    QueryPerformanceCounter(&now);
  } while (now.QuadPart < expire.QuadPart);
}

#endif /* defined(ADMXRC3_IMPORT_DELAY_US) */

#endif
